#include <math.h>
#include "winbase.h"
#include "wingdi.h"
#include "winerror.h"
#include "wine/debug.h"

WINE_DEFAULT_DEBUG_CHANNEL(msimg32);

BOOL WINAPI GradientFill(HDC hdc, PTRIVERTEX pVertex, ULONG nVertex, 
                         PVOID pMesh, ULONG dwNumMesh, ULONG dwMode )
{
  FIXME("stub: %ld vertices %ld gradients mode %lx\n", nVertex, dwNumMesh, dwMode);
  SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
  return FALSE;
}

BOOL WINAPI TransparentBlt(HDC hdcDest,
			   int nXOriginDest, int nYOriginDest,
			   int nWidthDest, int nHeightDest,
			   HDC hdcSrc,
			   int nXOriginSrc, int nYOriginSrc,
			   int nWidthSrc, int nHeightSrc,
			   UINT crTransparent)
{
  COLORREF bkcol;
  HBITMAP hbmMask, hbmImage, hbmTemp;
  HDC hdcMask, hdcImage, hdcTemp;

  TRACE("(%d, (%d,%d),(%d,%d), %d, (%d,%d),(%d,%d), %d) - not optimized\n",
        hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest,
	hdcSrc, nXOriginSrc, nYOriginSrc, nWidthSrc, nHeightSrc, crTransparent);

  /* FIXME: we do not support other map modes than MM_TEXT */
  /* FIXME: we're also supposed to do something about the StretchBltMode */

  /* create monochrome bitmap for the transparent pixels */
  hbmMask = CreateBitmap(nWidthSrc, nHeightSrc, 1, 1, NULL);
  hdcMask = CreateCompatibleDC(hdcDest);
  SelectObject(hdcMask, hbmMask);
  bkcol = SetBkColor(hdcSrc, crTransparent);
  BitBlt(hdcMask, 0, 0, nWidthSrc, nHeightSrc, hdcSrc, nXOriginSrc, nYOriginSrc, SRCCOPY);
  SetBkColor(hdcSrc, bkcol);

  /* create color bitmap for the source image */
  hbmImage = CreateCompatibleBitmap(hdcDest, nWidthSrc, nHeightSrc);
  hdcImage = CreateCompatibleDC(hdcDest);
  SelectObject(hdcImage, hbmImage);
  BitBlt(hdcImage, 0, 0, nWidthSrc, nHeightSrc, hdcSrc, nXOriginSrc, nYOriginSrc, SRCCOPY);
  /* remove color key from it */
  BitBlt(hdcImage, 0, 0, nWidthSrc, nHeightSrc, hdcMask, 0, 0, 0x220326);

#if 1
  /* The two StretchBlts could be done directly on the destination, but to avoid
   * flicker, we'll copy the destination and work on that instead. This is slower,
   * but easier on the eyes. */
  hbmTemp = CreateCompatibleBitmap(hdcDest, nWidthDest, nHeightDest);
  hdcTemp = CreateCompatibleDC(hdcDest);
  SelectObject(hdcTemp, hbmTemp);
  BitBlt(hdcTemp, 0, 0, nWidthDest, nHeightDest, hdcDest, nXOriginDest, nYOriginDest, SRCCOPY);

  /* clear the destination parts we don't want */
  StretchBlt(hdcTemp, 0, 0, nWidthDest, nHeightDest,
             hdcMask, 0, 0, nWidthSrc, nHeightSrc,
             SRCAND);
  /* fill them with the source parts we want */
  StretchBlt(hdcTemp, 0, 0, nWidthDest, nHeightDest,
             hdcImage, 0, 0, nWidthSrc, nHeightSrc,
             SRCPAINT);

  /* blit the modifications back to the real destination */
  BitBlt(hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest, hdcTemp, 0, 0, SRCCOPY);

  DeleteDC(hdcTemp);
  DeleteObject(hbmTemp);
#else
  /* flickerful version */

  /* clear the destination parts we don't want */
  StretchBlt(hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest,
             hdcMask, 0, 0, nWidthSrc, nHeightSrc,
             SRCAND);
  /* fill them with the source parts we want */
  StretchBlt(hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest,
             hdcImage, 0, 0, nWidthSrc, nHeightSrc,
             SRCPAINT);
#endif

  DeleteDC(hdcImage);
  DeleteObject(hbmImage);
  DeleteDC(hdcMask);
  DeleteObject(hbmMask);

  return TRUE;
}

BOOL WINAPI AlphaBlend(HDC hdcDest,
                       int nXOriginDest, int nYOriginDest,
                       int nWidthDest, int nHeightDest,
		       HDC hdcSrc,
		       int nXOriginSrc, int nYOriginSrc,
		       int nWidthSrc, int nHeightSrc,
		       BLENDFUNCTION blendFunction)
{
  HDC hdcScratch;
  BITMAPINFO bmiScratch;
  HBITMAP hbmSrc, hbmDest;
  BYTE *src, *dest;
  int x,y,c;
  double scaleX, scaleY;
  
  TRACE("(%d, (%d,%d),(%d,%d), %d, (%d,%d),(%d,%d), (%d,%d,%d,%d)) stub!\n",
        hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest,
	hdcSrc, nXOriginSrc, nYOriginSrc, nWidthSrc, nHeightSrc, 
	blendFunction.BlendOp, blendFunction.BlendFlags, 
	blendFunction.SourceConstantAlpha, blendFunction.AlphaFormat);
  
  if ((blendFunction.BlendOp != AC_SRC_OVER) || blendFunction.BlendFlags) {
    FIXME("Invalid parameters\n");
    return FALSE;
  }
  
  /* Check for opaque case that doesn't need alpha blending */
  if ((blendFunction.SourceConstantAlpha == 255) && (blendFunction.AlphaFormat == 0))
  {
    StretchBlt(hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest,
               hdcSrc, nXOriginSrc, nYOriginSrc, nWidthSrc, nHeightSrc, SRCCOPY);
    return TRUE;
  }
  
  /* create scratch dc and prep the dibsection header */
  /* FIXME: change bmiScratch to a BITMAPV5HEADER to ensure DIB sections have alpha */
  memset(&bmiScratch.bmiHeader, 0, sizeof(BITMAPINFOHEADER));
  bmiScratch.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmiScratch.bmiHeader.biWidth = nWidthSrc;
  bmiScratch.bmiHeader.biHeight = nHeightSrc;
  bmiScratch.bmiHeader.biPlanes = 1;
  bmiScratch.bmiHeader.biBitCount = 32;
  bmiScratch.bmiHeader.biCompression = BI_RGB;
  
  hdcScratch = CreateCompatibleDC(hdcDest);
  
  /* read in src */
  hbmSrc = CreateDIBSection(hdcScratch, &bmiScratch, DIB_RGB_COLORS, (void**)(&src), 0, 0);
  SelectObject(hdcScratch, hbmSrc);

  /* Using BitBlt will preserve the alpha in the original DIB (if there is one), since we do a dib->dib copy */
  BitBlt(hdcScratch, 0, 0, nWidthSrc, nHeightSrc, hdcSrc, nXOriginSrc, nYOriginSrc, SRCCOPY);
  
  /* read in dest */
  bmiScratch.bmiHeader.biWidth = nWidthDest;
  bmiScratch.bmiHeader.biHeight = nHeightDest;
  hbmDest = CreateDIBSection(hdcScratch, &bmiScratch, DIB_RGB_COLORS, (void**)(&dest), 0, 0);
  SelectObject(hdcScratch, hbmDest);
  BitBlt(hdcScratch, 0, 0, nWidthDest, nHeightDest, hdcDest, nXOriginDest, nYOriginDest, SRCCOPY);

  /* Calculate the scaling factor for sampling - note that we're doing a trivial point-sample on 
   * the source data */
  scaleX = (double)nWidthSrc / (double)nWidthDest;
  scaleY =(double)nHeightSrc / (double)nHeightDest;

  /* blend them together in dest */
  if (blendFunction.AlphaFormat == AC_SRC_ALPHA)
  {
    FLOAT sca = blendFunction.SourceConstantAlpha / 255.0;
    for (y=0; y<nHeightDest; y++) {
      for (x=0; x<nWidthDest; x++) {
        FLOAT sa = src[((int)floor(y * scaleY) * 4 * nWidthSrc) + ((int)floor(x * scaleX) * 4) + 3] / 255.0;
        for (c=0; c<4; c++)
        {
          BYTE s = src[((int)floor(y * scaleY) * 4 * nWidthSrc) + ((int)floor(x * scaleX) * 4) + c];
          BYTE d = dest[(y * 4 * nWidthDest) + (x * 4) + c];
          FLOAT t = (s * sca) + ((1 - sa) * d);
          dest[(y * 4 * nWidthDest) + (x * 4) + c] = max(0, min(t, 255));
        }
      }
    }
  } else {
    FLOAT sca = blendFunction.SourceConstantAlpha / 255.0;
    for (y=0; y<nHeightDest; y++) {
      for (x=0; x<nWidthDest; x++) {
        for (c=0; c<4; c++)
        {
          BYTE s = src[((int)floor(y * scaleY) * 4 * nWidthSrc) + ((int)floor(x * scaleX) * 4) + c];
          BYTE d = dest[(y * 4 * nWidthDest) + (x * 4) + c];
          FLOAT t = (s * sca) + ((1 - sca) * d);
          dest[(y * 4 * nWidthDest) + (x * 4) + c] = max(0, min(t, 255));
        }
      }
    }
  }
  
  /* copy to destination dc */
  BitBlt(hdcDest, nXOriginDest, nYOriginDest, nWidthDest, nHeightDest, hdcScratch, 0, 0, SRCCOPY);
  
  /* delete objects */
  DeleteObject(hbmDest);
  DeleteObject(hbmSrc);
  DeleteObject(hdcScratch);
  
  return TRUE;
}
