/*
 * Implementation of InternetShortcut class
 *
 * Copyright (c) 2005 TransGaming Technologies
 */

#include "wine/debug.h"
#include "shdocvw.h"
#include "oleauto.h"
#include "wine/unicode.h"

WINE_DEFAULT_DEBUG_CHANNEL(shdocvw);

/**********************************************************************
 * Implement the InternetShortcut class
 */

static ICOM_VTABLE(IUniformResourceLocatorA) IS_Vtbl;
static ICOM_VTABLE(IUniformResourceLocatorW) ISW_Vtbl;
static ICOM_VTABLE(IPersistFile) ISPF_Vtbl;

static HRESULT WINAPI IS_QueryInterface(LPUNIFORMRESOURCELOCATORA iface, REFIID riid, LPVOID *ppobj)
{
    ICOM_THIS(InternetShortcutImpl, iface);

    TRACE("(%p)->(%s, %p)\n", iface, debugstr_guid(riid), ppobj);
    if (IsEqualGUID(&IID_IUnknown, riid) ||
        IsEqualGUID(&IID_IUniformResourceLocatorA, riid))
    {
        This->ref++;
        *ppobj = This;
        TRACE("IUnknown or IUniformResourceLocatorA\n");
        return S_OK;
    }
    else if (IsEqualGUID(&IID_IUniformResourceLocatorW, riid))
    {
        This->ref++;
        *ppobj = (LPVOID *)&This->lpVtblW;
        TRACE("IUniformResourceLocator\n");
        return S_OK;
    }
    else if (IsEqualGUID(&IID_IPersistFile, riid))
    {
        This->ref++;
        *ppobj = (LPVOID *)&This->lpVtbl_PF;
        TRACE("IPersistFile\n");
        return S_OK;
    }
    else
    {
        FIXME("(%p)->(%s, %p): no interface\n", iface, debugstr_guid(riid), ppobj);
        return E_NOINTERFACE;
    }
}

static HRESULT WINAPI ISW_QueryInterface(LPUNIFORMRESOURCELOCATORW iface, REFIID riid, LPVOID *ppobj)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);
    
    TRACE("(%p)->(%s, %p)\n", iface, debugstr_guid(riid), ppobj);
    return IUniformResourceLocatorA_QueryInterface((LPUNIFORMRESOURCELOCATORA) This, riid, ppobj);
}

static HRESULT WINAPI ISPF_QueryInterface(LPPERSISTFILE iface, REFIID riid, LPVOID *ppobj)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);
    
    TRACE("(%p)->(%s, %p)\n", iface, debugstr_guid(riid), ppobj);
    return IUniformResourceLocatorA_QueryInterface((LPUNIFORMRESOURCELOCATORA) This, riid, ppobj);
}

static ULONG WINAPI IS_AddRef(LPUNIFORMRESOURCELOCATORA iface)
{
    ICOM_THIS(InternetShortcutImpl, iface);
    TRACE("(%p)->()\n", iface);
    return ++(This->ref);
}

static ULONG WINAPI ISW_AddRef(LPUNIFORMRESOURCELOCATORW iface)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);
    
    TRACE("(%p)->()\n", iface);
    return IUniformResourceLocatorA_AddRef((LPUNIFORMRESOURCELOCATORA) This);
}

static ULONG WINAPI ISPF_AddRef(LPPERSISTFILE iface)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);
    
    TRACE("(%p)->()\n", iface);
    return IUniformResourceLocatorA_AddRef((LPUNIFORMRESOURCELOCATORA) This);
}

static ULONG WINAPI IS_Release(LPUNIFORMRESOURCELOCATORA iface)
{
    ICOM_THIS(InternetShortcutImpl, iface);
    TRACE("(%p)->()\n", iface);
    if (--(This->ref)) return This->ref;

    HeapFree(GetProcessHeap(), 0, This);
    return 0;
}

static ULONG WINAPI ISW_Release(LPUNIFORMRESOURCELOCATORW iface)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);
    
    TRACE("(%p)->()\n", iface);
    return IUniformResourceLocatorA_Release((LPUNIFORMRESOURCELOCATORA) This);
}

static ULONG WINAPI ISPF_Release(LPPERSISTFILE iface)
{
    ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);
    
    TRACE("(%p)->()\n", iface);
    return IUniformResourceLocatorA_Release((LPUNIFORMRESOURCELOCATORA) This);
}

static HRESULT WINAPI IS_SetURL(LPUNIFORMRESOURCELOCATORA iface, LPCSTR pcstrURL, DWORD dwFlags)
{
    /*ICOM_THIS(InternetShortcutImpl, iface);*/
    FIXME("(%p)->(%s,0x%lx): stub\n", iface, debugstr_a(pcstrURL), dwFlags);
    return S_OK;
}

static HRESULT WINAPI ISW_SetURL(LPUNIFORMRESOURCELOCATORW iface, LPCWSTR pcstrURL, DWORD dwFlags)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);*/
    FIXME("(%p)->(%s,0x%lx): stub\n", iface, debugstr_w(pcstrURL), dwFlags);
    return S_OK;
}

static HRESULT WINAPI IS_GetURL(LPUNIFORMRESOURCELOCATORA iface, LPSTR* pstrURL)
{
    /*ICOM_THIS(InternetShortcutImpl, iface);*/
    FIXME("(%p)->(%p): stub\n", iface, pstrURL);
    *pstrURL = 0;
    return S_OK;
}

static HRESULT WINAPI ISW_GetURL(LPUNIFORMRESOURCELOCATORW iface, LPWSTR* pstrURL)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);*/
    FIXME("(%p)->(%p): stub\n", iface, pstrURL);
    *pstrURL = 0;
    return S_OK;
}

static HRESULT WINAPI IS_InvokeCommand(LPUNIFORMRESOURCELOCATORA iface, PURLINVOKECOMMANDINFOA pInv)
{
    /*ICOM_THIS(InternetShortcutImpl, iface);*/
    FIXME("(%p)->(%p): stub\n", iface, pInv);
    return S_OK;
}

static HRESULT WINAPI ISW_InvokeCommand(LPUNIFORMRESOURCELOCATORW iface, PURLINVOKECOMMANDINFOW pInv)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtblW, iface);*/
    FIXME("(%p)->(%p): stub\n", iface, pInv);
    return S_OK;
}

static HRESULT WINAPI ISPF_GetClassID(LPPERSISTFILE iface, CLSID* pClassID)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    TRACE("(%p)->(%p)\n", iface, pClassID);
    memcpy(pClassID, &CLSID_InternetShortcut, sizeof(CLSID));
    return S_OK;
}

static HRESULT WINAPI ISPF_IsDirty(LPPERSISTFILE iface)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    FIXME("(%p)->(): stub\n", iface);
    return S_OK;
}

static HRESULT WINAPI ISPF_Load(LPPERSISTFILE iface, LPCOLESTR pszFileName, DWORD dwMode)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    FIXME("(%p)->(%s,0x%lx): stub\n", iface, debugstr_w(pszFileName), dwMode);
    return S_OK;
}

static HRESULT WINAPI ISPF_Save(LPPERSISTFILE iface, LPCOLESTR pszFileName, BOOL fRemember)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    FIXME("(%p)->(%s,%d): stub\n", iface, debugstr_w(pszFileName), fRemember);
    return S_OK;
}

static HRESULT WINAPI ISPF_SaveCompleted(LPPERSISTFILE iface, LPCOLESTR pszFileName)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    FIXME("(%p)->(%s): stub\n", iface, debugstr_w(pszFileName));
    return S_OK;
}

static HRESULT WINAPI ISPF_GetCurFile(LPPERSISTFILE iface, LPOLESTR* ppszFileName)
{
    /*ICOM_THIS_MULTI(InternetShortcutImpl, lpVtbl_PF, iface);*/
    FIXME("(%p)->(%p): stub\n", iface, ppszFileName);
    *ppszFileName = 0;
    return S_OK;
}

HRESULT SHDOCVW_CreateShortcut(IUnknown *pOuter, REFIID riid, LPVOID *ppobj)
{
    InternetShortcutImpl *ipIS;
    HRESULT hr;
    TRACE("()\n");

    ipIS = (InternetShortcutImpl *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                                             sizeof(InternetShortcutImpl));
    if (ipIS == NULL) return E_OUTOFMEMORY;
    ICOM_VTBL(ipIS) = &IS_Vtbl;
    ipIS->lpVtblW = &ISW_Vtbl;
    ipIS->lpVtbl_PF = &ISPF_Vtbl;

    IUniformResourceLocatorA_AddRef((IUniformResourceLocatorA *)ipIS);
    hr = IUniformResourceLocatorA_QueryInterface((IUniformResourceLocatorA *)ipIS, riid, ppobj);
    IUniformResourceLocatorA_Release((IUniformResourceLocatorA *)ipIS);

    if (SUCCEEDED(hr))
        return S_OK;
    else
        return E_NOINTERFACE;
}

/**********************************************************************
 * IUniformResourceLocatorA virtual function table
 */

static ICOM_VTABLE(IUniformResourceLocatorA) IS_Vtbl =
{
    ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
    IS_QueryInterface,
    IS_AddRef,
    IS_Release,
    IS_SetURL,
    IS_GetURL,
    IS_InvokeCommand
};

/**********************************************************************
 * IUniformResourceLocatorW virtual function table
 */

static ICOM_VTABLE(IUniformResourceLocatorW) ISW_Vtbl =
{
    ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
    ISW_QueryInterface,
    ISW_AddRef,
    ISW_Release,
    ISW_SetURL,
    ISW_GetURL,
    ISW_InvokeCommand
};

/**********************************************************************
 * IPersistFile virtual function table
 */

static ICOM_VTABLE(IPersistFile) ISPF_Vtbl =
{
    ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
    ISPF_QueryInterface,
    ISPF_AddRef,
    ISPF_Release,
    ISPF_GetClassID,
    ISPF_IsDirty,
    ISPF_Load,
    ISPF_Save,
    ISPF_SaveCompleted,
    ISPF_GetCurFile
};
